# uCTest

The uCTest framework is especially developed for microcontrollers. You are able to define tests and run them automatically and protocol the test results.

## Features

- Testsuites
- Tests
- Test Fixtures
  - Setup
  - Teardown
- Reports
  - Text (Like GTest)
  - JSON (convert via python - JUnit XML)
  - XML (AVR via Runner - JUnit XML)
  - None
- System Setup
- System Teardown

## Install

uCTest is configuralbe at compile time. For this some options are valid only for some architectures.

| Flag                    | Architecture       | Description                                |
| ------------------------| ------------------ | ------------------------------------------ |
| CONFIG_USART            | avr                | Output via USART                           |
| CONFIG_USART_BAUD       | avr                | USART Baudrate                             |
| CONFIG_COUNT_CYCLES     | avr                | Cycle counter for performance measurement  |
| CONFIG_OUTPUT_NONE      | avr, riscv, x86_64 | No output                                  |
| CONFIG_OUTPUT_SIMPLE    | avr, riscv, x86_64 | Test output like GTest                     |
| CONFIG_OUTPUT_JSON      | avr, riscv, x86_64 | json output. Convert with python to XML    |
| CONFIG_OUTPUT_SIMULATOR | avr                | The runner makes the output (XML and Text) |

It is possible to append the options with the makefile call. For the output is only one valid in the list of ``NONE``, ``SIMPLE``, ``JSON`` or ``SIMULATOR``. And if you define a baudrate the ``USART`` option is implicit set.

```sh
$ make ARCH=avr MCU=atmega328p F_CPU=11059200UL OUTPUT=SIMPLE BAUD=115200UL
  CC      src/System_avr.o
  CC      src/uCTest.o
  CC      src/messages.o
  AR      uCTest.a
/usr/bin/avr-ar: creating libuCTest.a
  RANLIB  uCTest.a
  HOSTCC  simulator/main.o
  HOSTCC  simulator/test.o
  LINK    host/testRunner
  LINK    host/libtestRunner.so
$ make install INSTALL_DIR=/project/uctest
```

Here we compile the test for the AVR ATmega328P. We use the normal text as the output option and comunicates with the USART interface (115200Baud 8N1). At the end we install uCTest.

## A simple test

```c
#include <uCTest/uCTest.h>

TESTSUITE( SimpleTest );

TEST( SimpleTest, Success ) {
    EXPECT_EQ( 1, 1 );
    EXPECT_EQ( 2, 2 );
    EXPECT_EQ( 3, 3 );
}

TEST( SimpleTest, Fail ) {
    EXPECT_NE( 1, 1 );
}
```

```sh
$ avr-gcc -Iinclude -I/project/uctest/include -mmcu=atmega328p -DF_CPU=11059200UL -c -o test.o test.c
$ avr-gcc -mmcu=atmega328p -Wl,-T/project/uctest/lib/ldscripts/avr/avr5.xn -Wl,-u,vfprintf -lprintf_min -o test.elf test.o -L/project/uctest/lib/ -luCTest
$ ./host/testRunner -m atmega328p ./test.elf
Loaded 5392 .text at address 0x0
Loaded 42 .data
Running main
[==========] Running 2 tests from 1 test suites.
[----------] 2 tests from SimpleTest
[ RUN      ] SimpleTest.Success
[       OK ] SimpleTest.Success (58 cycles)
[ RUN      ] SimpleTest.Fail
src/test.c:12: Failure
Expected: ( 1 ) != ( 1 ), actual: 1 vs 1
[  FAILED  ] SimpleTest.Fail (169 cycles)
[----------] 2 tests from SimpleTest (227 cycles)

[==========] 2 tests from 1 test suites ran. (227 cycles total)
[  PASSED  ] 1 tests.
[  FAILED  ] 1 tests, listed below:
[  FAILED  ] SimpleTest.Fail (0 cycles)

1 FAILED TESTS
```

This example runs with the help of the Runner. But is is runnable directly on a microcontroller. It uses 5kB of the flash storage.

It is possible to run this test suite also on a RISC-V controller. For this we have to re-translate the uCTest for the right architecture.

```sh
$ make ARCH=riscv
  CC      src/System_riscv.o
  CC      src/uCTest.o
  CC      src/messages.o
  AR      uCTest.a
/usr/bin/avr-ar: creating libuCTest.a
  RANLIB  uCTest.a
$ make install INSTALL_DIR=/project/uctest
$ /usr/bin/riscv32-unknown-elf-gcc -I/project/uctest/include -march=rv32imac -mabi=ilp32 -c -o test.o src/test.c
$ /usr/bin/riscv32-unknown-elf-gcc -march=rv32imac -mabi=ilp32 -specs=/project/uctest/lib/ldscripts/riscv/picolibc.specs -Wl,-T/project/uctest/lib/ldscripts/riscv/sifive-e31.ld --oslib=semihost --crt0=hosted -o test.elf test.o -L/project/uctest/lib -luCTest
$ qemu-system-riscv32 -chardev stdio,mux=on,id=stdio0 -semihosting-config enable=on,chardev=stdio0 -mon chardev=stdio0,mode=readline -serial chardev:stdio0 -machine sifive_e,revb=true -cpu sifive-e31 -kernel test.elf -bios none -nographic
Running main
[==========] Running 2 tests from 1 test suites.
[----------] 2 tests from SimpleTest
[ RUN      ] SimpleTest.Success
[       OK ] SimpleTest.Success (0 cycles)
[ RUN      ] SimpleTest.Fail
src/test.c:12: Failure
Expected: ( 1 ) != ( 1 ), actual: 1 vs 1
[  FAILED  ] SimpleTest.Fail (0 cycles)
[----------] 2 tests from SimpleTest (0 cycles)

[==========] 2 tests from 1 test suites ran. (0 cycles total)
[  PASSED  ] 1 tests.
[  FAILED  ] 1 tests, listed below:
[  FAILED  ] SimpleTest.Fail (0 cycles)

1 FAILED TESTS
```
