/**
 * @file
 * @defgroup   uCTest uCTest: Unit testing framework for microcontrollers
 * @brief      This is a unit testing framework for microcontrollers
 *
 * A simple unit test framework for microcontrollers. It is developed first on AVR devices.
 *
 * @author     Sebastian Koschmieder <sep@seplog.org>
 * @date       2021-01-12 21:04:41
 * @{
 */
#include "uCTest_device.h"
#include "uCTest_macro.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _UCTEST_UCTEST_H_
#if !defined( __DOXYGEN__ )
#define _UCTEST_UCTEST_H_
#endif

#if !defined( __DOXYGEN__ )
enum uctest_tst_flags {
    UCTEST_TEST_NONE,
    UCTEST_TEST_SIMPLE,
    UCTEST_TEST_FIXTURE,
    UCTEST_TEST_SETUP,
    UCTEST_TEST_TEARDOWN
};

struct uctest_test {
    const __flash char *suite_name;
    const __flash char *test_name;
    enum uctest_tst_flags tst_flags;
    const __flash char *file;
    uint16_t line;
    void ( *test_func )( void );
};

typedef void ( *uctest_register )( const __flash struct uctest_test ** );
#define uctest_register_prototype( name ) \
    void __attribute__( ( used ) ) name( const __flash struct uctest_test **tst )

#define TEST_TEMPLATE( suite, name, flags, ... )                                          \
    uctest_register_prototype( suite##_##name##_register ) {                              \
        extern void _##suite##_##name( void );                                            \
        static const __flash char suite##_##name##_name[] = UCTEST_STR( name );           \
        static const __flash struct uctest_test suite##_##name##_test = {                 \
            .suite_name = ( const __flash char * ) &( suite ),                            \
            .test_name = ( const __flash char * ) &( suite##_##name##_name ),             \
            .tst_flags = flags,                                                           \
            .file = _##suite##_file,                                                      \
            .line = __LINE__,                                                             \
            .test_func = ( _##suite##_##name ) };                                         \
        *tst = ( const __flash struct uctest_test * ) &suite##_##name##_test;             \
        UCTEST_FOREACH( UCTEST_CHECK_FUNCTION_PRESENT, ##__VA_ARGS__ );                   \
    }                                                                                     \
    static const uctest_register _##suite##_##name##_ptr                                  \
        __attribute__( ( section( ".uctest.list" ), used ) ) = suite##_##name##_register; \
    void _##suite##_##name( void )

#define _EXPECT_SIMPLE( WHAT, CMP, a, b )                                             \
    do {                                                                              \
        if( !( CMP( a, b ) ) ) {                                                      \
            extern const __flash char uctest_simpleFailMsg[];                         \
            WHAT( uctest_simpleFailMsg, UCTEST_STR( CMP( a, b ) ), ( ( int ) ( a ) ), \
                  ( ( int ) ( b ) ) );                                                \
        }                                                                             \
    } while( 0 )
#endif

#if defined( __DOXYGEN__ )
/**
 * @defgroup test Test: Test definitions
 * @{
 */
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Define a testsuite
 *
 * The tests are grouped within a testsuite. To define a testsuite use this macro. This will create
 * a variable with the given name to use with the tests.
 *
 * @param[in]  name of the suite
 */
#define TESTSUITE( name )
#else
#define TESTSUITE( name )                           \
    const __flash char name[] = UCTEST_STR( name ); \
    const __flash char _##name##_file[] = __FILE__
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Define a single test for the testsuite
 *
 * To define a test use this macro. The tests are not ordered. All of the defined tests will run
 * within the testsuite. For every test it is possible to log the cycles they used to compute.
 *
 * @see TESTSUITE
 *
 * @param[in]  suite The name of the suite
 * @param[in]  name The name of the test
 */
#define TEST( suite, name )
#else
#define TEST( suite, name ) TEST_TEMPLATE( suite, name, UCTEST_TEST_SIMPLE )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Define a test fixture for the testsuite
 *
 * With this macro you define a test fixture for the testsuite. The aim of the test fixture is, to
 * have always the same conditions of a test. For this you have to use the SETUP and the TEARDOWN
 * macro.
 *
 * @see TESTSUITE
 * @see SETUP
 * @see TEARDOWN
 *
 * @param[in]  suite The name of the suite
 * @param[in]  name The name of the test
 */
#define TEST_F( suite, name )
#else
#define TEST_F( suite, name ) \
    TEST_TEMPLATE( suite, name, UCTEST_TEST_FIXTURE, _##suite##_SetUp, _##suite##_TearDown )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Defines a setup function for a test fixture within a testsuite
 *
 * Define a setup for every test fixture within a testsuite. Before every run of a test fixture
 * this setup will be exected to enshure always a valid state.
 *
 * @see TESTSUITE
 * @see TEST_F
 * @see TEARDOWN
 *
 * @param[in] suite Name of the testsuite
 */
#define SETUP( suite )
#else
#define SETUP( suite ) TEST_TEMPLATE( suite, SetUp, UCTEST_TEST_SETUP )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Define a teardown function for the test fixture within a testsuite
 *
 * This defines the teardown for the test fixture. After the execution of the test function, this
 * teardown will be executed.
 *
 * @see TESTSUITE
 * @see TEST_F
 * @see SETUP
 *
 * @param[in] suite The name of the suite
 */
#define TEARDOWN( suite )
#else
#define TEARDOWN( suite ) TEST_TEMPLATE( suite, TearDown, UCTEST_TEST_TEARDOWN )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Run all registered tests
 * @return an exit code
 */
#define RUN_ALL_TESTS()
#else
#define RUN_ALL_TESTS()                   \
    ( {                                   \
        extern int run_all_tests( void ); \
        run_all_tests();                  \
    } )
#endif

/** @} */

/**
 * @defgroup assertions Assertions: Assertions to use in tests
 *
 * The assertions are ordered in assertions which generate an error and abort the execution (starts
 * with ASSERT) and assertions wich generate an error and continue the execution (starts with
 * EXPECT)
 *
 * @{
 */

#if defined( __DOXYGEN__ )
/**
 * @brief Fail the test and abort the execution
 *
 * This fails the test and will abort the execution. It is possible to add a user defined message.
 * For this is use the first parameter as the format string and the other variable aguments as
 * arguments like you use for printf
 *
 * @see ADD_FAILURE
 *
 * @param ... The first have to be the format string. After this, the arguments like printf
 */
#define FAIL( ... )
#else
#define UCTEST_FAIL_DEFAULT( file, line )                                                   \
    do {                                                                                    \
        extern void uctest_testFail( const __flash char *, int, const __flash char *, ... ) \
            __attribute__( ( format( printf, 3, 4 ) ) );                                    \
        uctest_testFail( file, line, ( ( void * ) 0 ) );                                    \
        return;                                                                             \
    } while( 0 )

#define UCTEST_FAIL_ARGS( file, line, fmt, ... )                                                \
    do {                                                                                        \
        extern void uctest_testFail( const __flash char *, int, const __flash char *, ... )     \
            __attribute__( ( format( printf, 3, 4 ) ) );                                        \
        void check_args( const char *format, ... ) __attribute__( ( format( printf, 1, 2 ) ) ); \
        if( 0 )                                                                                 \
            check_args( fmt, ##__VA_ARGS__ );                                                   \
        uctest_testFail( file, line, fmt, ##__VA_ARGS__ );                                      \
        return;                                                                                 \
    } while( 0 )

#define UCTEST_FAIL( file, line, ... )                                          \
    UCTEST_CALL_DEFARGS( UCTEST_FAIL_DEFAULT, UCTEST_FAIL_ARGS, ##__VA_ARGS__ ) \
    ( file, line, ##__VA_ARGS__ )

#define FAIL( ... ) UCTEST_FAIL( __FILE__, __LINE__, ##__VA_ARGS__ )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief This adds a failure for the test execution but the test is not aborted
 *
 * This is nearly the same as the FAIL macro. Except that the execution will not be aborted. The
 * execution of the test will continue. It is also possible to define a user defined message.
 *
 * @see FAIL
 *
 * @param ... The first argument is the format string. After this the parameters like printf
 */
#define ADD_FAILURE( ... )
#else
#define UCTEST_ADD_FAILURE_DEFAULT( file, line )                                            \
    do {                                                                                    \
        extern void uctest_testFail( const __flash char *, int, const __flash char *, ... ) \
            __attribute__( ( format( printf, 3, 4 ) ) );                                    \
        uctest_testFail( file, line, ( ( void * ) 0 ) );                                    \
    } while( 0 )

#define UCTEST_ADD_FAILURE_ARGS( file, line, fmt, ... )                                         \
    do {                                                                                        \
        extern void uctest_testFail( const __flash char *, int, const __flash char *, ... )     \
            __attribute__( ( format( printf, 3, 4 ) ) );                                        \
        void check_args( const char *format, ... ) __attribute__( ( format( printf, 1, 2 ) ) ); \
        if( 0 )                                                                                 \
            check_args( fmt, ##__VA_ARGS__ );                                                   \
        uctest_testFail( file, line, fmt, ##__VA_ARGS__ );                                      \
    } while( 0 )

#define UCTEST_ADD_FAILURE( file, line, ... )                                                   \
    UCTEST_CALL_DEFARGS( UCTEST_ADD_FAILURE_DEFAULT, UCTEST_ADD_FAILURE_ARGS, ##__VA_ARGS__ \ ) \
    ( file, line, ##__VA_ARGS__ )

#define ADD_FAILURE( ... ) UCTEST_ADD_FAILURE( __FILE__, __LINE__, ##__VA_ARGS__ )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check equality
 *
 * Check the equality and register an error if the values are not equal
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_EQ( a, b )
/**
 * @brief Check equality
 *
 * Check the equality and abort the test if the values are not equal
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define ASSERT_EQ( a, b )
#else
#define _EXPECT_EQ( WHAT, a, b )                                                            \
    do {                                                                                    \
        if( ( a ) != ( b ) ) {                                                              \
            extern const __flash char uctest_eqFailMsg[];                                   \
            WHAT( uctest_eqFailMsg, ( #a ), ( ( int ) ( a ) ), ( #b ), ( ( int ) ( b ) ) ); \
        }                                                                                   \
    } while( 0 )

#define EXPECT_EQ( a, b ) _EXPECT_EQ( ADD_FAILURE, a, b )

#define ASSERT_EQ( a, b ) _EXPECT_EQ( FAIL, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for no equality
 *
 * Check for no equality. If the values are equal register an error.
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_NE( a, b )
/**
 * @brief Check for no equality
 *
 * Check for no equality. If the values are equal abort the test
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define ASSERT_NE( a, b )
#else
#define _CMP_NE( a, b ) ( a ) != ( b )
#define EXPECT_NE( a, b ) _EXPECT_SIMPLE( ADD_FAILURE, _CMP_NE, a, b )
#define ASSERT_NE( a, b ) _EXPECT_SIMPLE( FAIL, _CMP_NE, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for less than
 *
 * Check for less than. If the test fail, register an error
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_LT( a, b )
/**
 * @brief Check for less than
 *
 * Check for less than. If the test fail, abort the test
 *
 * @param[in] a current
 * @param[in] b expected
 */

#define ASSERT_LT( a, b )
#else
#define _CMP_LT( a, b ) ( a ) < ( b )
#define EXPECT_LT( a, b ) _EXPECT_SIMPLE( ADD_FAILURE, _CMP_LT, a, b )
#define ASSERT_LT( a, b ) _EXPECT_SIMPLE( FAIL, _CMP_LT, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for less or equal
 *
 * Check for less or equal. If the test fail, register an error
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_LE( a, b )
/**
 * @brief Check for less or equal
 *
 * Check for less or equal. If the test fail, abort the test
 *
 * @param[in] a current
 * @param[in] b expected
 */

#define ASSERT_LE( a, b )
#else
#define _CMP_LE( a, b ) ( a ) <= ( b )
#define EXPECT_LE( a, b ) _EXPECT_SIMPLE( ADD_FAILURE, _CMP_LE, a, b )
#define ASSERT_LE( a, b ) _EXPECT_SIMPLE( FAIL, _CMP_LE, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for greater than
 *
 * Check for greater than. If the test fail, register an error
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_GT( a, b )
/**
 * @brief Check for greater than
 *
 * Check for greater than. If the test fail, abort the test
 *
 * @param[in] a current
 * @param[in] b expected
 */

#define ASSERT_GT( a, b )
#else
#define _CMP_GT( a, b ) ( a ) > ( b )
#define EXPECT_GT( a, b ) _EXPECT_SIMPLE( ADD_FAILURE, _CMP_GT, a, b )
#define ASSERT_GT( a, b ) _EXPECT_SIMPLE( FAIL, _CMP_GT, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for greater or equal
 *
 * Check for greater or equal. If the test fail, register an error
 *
 * @param[in] a current
 * @param[in] b expected
 */
#define EXPECT_GE( a, b )
/**
 * @brief Check for greater or equal
 *
 * Check for greater or equal. If the test fail, abort the test
 *
 * @param[in] a current
 * @param[in] b expected
 */

#define ASSERT_GE( a, b )
#else
#define _CMP_GE( a, b ) ( a ) >= ( b )
#define EXPECT_GE( a, b ) _EXPECT_SIMPLE( ADD_FAILURE, _CMP_GE, a, b )
#define ASSERT_GE( a, b ) _EXPECT_SIMPLE( FAIL, _CMP_GE, a, b )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check if the values are near to each other
 *
 * Calculates the difference of the both values. If the difference is bigger than the given diff,
 * register an error.
 *
 * @param[in] a current
 * @param[in] b expected
 * @param[in] diff the maximal difference
 */
#define EXPECT_NEAR( a, b, diff )
/**
 * @brief Check if the values are near to each other
 *
 * Calculates the difference of the both values. If the difference is bigger than the given diff,
 * abort the test execution.
 *
 * @param[in] a current
 * @param[in] b expected
 * @param[in] diff the maximal difference
 */
#define ASSERT_NEAR( a, b, diff )
#else
#define _EXPECT_NEAR( WHAT, a, b, c )                                                          \
    do {                                                                                       \
        __typeof__( a ) _div = abs( ( a ) - ( b ) );                                           \
        if( _div > abs( c ) ) {                                                                \
            extern const __flash char uctest_nearFailMsg[];                                    \
            WHAT( uctest_nearFailMsg, ( #a ), ( #b ), ( _div ), ( #c ), ( #a ), ( a ), ( #b ), \
                  ( b ), ( #c ), ( c ) );                                                      \
        }                                                                                      \
    } while( 0 )

#define EXPECT_NEAR( a, b, c ) _EXPECT_NEAR( ADD_FAILURE, a, b, c )

#define ASSERT_NEAR( a, b, c ) _EXPECT_NEAR( FAIL, a, b, c )
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief Check for a failed assert
 *
 * This checks for a failed assertion. If an assert failed, an std-abort will be called. Here we
 * check if the abort was called. If it was called the test was successful. Otherwise not.
 * To test the correct DEATH behavior you have to compile without NDEBUG! If the check failed,
 * We add an error and continue with the execution.
 *
 * TODO:
 * For this time now, the message will be ignored. But perhaps it is in respect to the memory
 * possible to implement this.
 *
 * @param[in] func The function to execute
 * @param[in] msg Until now it is ignored
 */
#define EXPECT_DEATH( func, msg )

/**
 * @brief Check for a failed assert
 *
 * @see EXPECT_DEATH
 * But here the execution will be aborted!
 *
 * @param[in] func The function to execute
 * @param[in] msg Until now it is ignored
 */
#define ASSERT_DEATH( func, msg )
#else
#define _EXPECT_DEATH( WHAT, name, func, msg )                                     \
    do {                                                                           \
        extern int uctest_assertDeath( void ( * )( void ), const __flash char * ); \
        void name( void ) {                                                        \
            func;                                                                  \
        }                                                                          \
        extern const __flash char uctest_deathFailMsg[];                           \
        if( !uctest_assertDeath( name, msg ) ) {                                   \
            WHAT( uctest_deathFailMsg, #func );                                    \
        }                                                                          \
    } while( 0 )

#define EXPECT_DEATH( func, msg )                                                   \
    _EXPECT_DEATH( ADD_FAILURE, UCTEST_CONCAT( __assert_wrapper_, __LINE__ ), func, \
                   UCTEST_XSTR( msg ) )

#define ASSERT_DEATH( func, msg ) \
    _EXPECT_DEATH( FAIL, UCTEST_CONCAT( __assert_wrapper_, __LINE__ ), func, UCTEST_XSTR( msg ) )
#endif

/** @} */

#endif /** _UCTEST_UCTEST_H_ */

#ifdef __cplusplus
}
#endif
/** @} */
