/**
 * @ingroup    uCTest
 * @defgroup   device Device: Device specific definitions
 *
 * This testframework should work on several processor arcitectures. For this it is necessary to
 * provide several device specific functions and types. These are some types you have to impelement
 * to support other hardware.
 *
 * To port this test framework it is essentually to understand how it works.
 * The tests are only function pointer to the test functions and a pointer to the suite name. This
 * structure has to pushed first to the stack. This must occure directly after the stack was
 * initialized. To register the tests the specific register function is used. If all tests was
 * pushed to the stack, the list will be closed with a NULL pointer.
 *
 * @author     Sebastian Koschmieder <sep@seplog.org>
 * @date       2021-01-23 19:27:58
 * @{
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _UCTEST_UCTEST_DEVICE_H_
#if !defined( __DOXYGEN__ )
#define _UCTEST_UCTEST_DEVICE_H_
#endif
#if defined( __AVR__ )
#include "device/uCTest_avr.h"
#elif defined( __riscv )
#include "device/uCTest_riscv.h"
#elif defined( __x86_64 )
#include "device/uCTest_x86_64.h"
#endif

#if defined( __DOXYGEN__ )
/**
 * @brief The type which is pushed to the stack
 *
 * This is the default type pushed onto the stack at once. For an 8 bit avr dievice this should be
 * a uint8_t
 */
typedef void stack_t;

/**
 * @brief Prototype of a function to push the test pointer onto the stack
 *
 * This is the prototype of the function to register the test function. This should be a naked
 * function and "called" before the main.
 *
 * @param[in]  name The name of the test register function
 */
#define uctest_register_prototype( name )

/**
 * @brief stores the complete context on the stack
 *
 * For the assert death it is necessary to store the complete context on the stack. If the function
 * we want to test returns as it is designed, everything is ok. But if it is aborted, we jump from
 * the abort back to the uctest_assertDeath. For this we have to restore the context!
 */
#define uctest_context_save()

/**
 * @brief restore the context from the stack
 *
 * @see uctest_context_save
 */
#define uctest_context_restore()

/**
 * @brief  Push a pointer to the stack
 *
 * This is used to push the pointer to the stack. This function should always inlined.
 *
 * @param[in] ptr Pointer to the test function
 */
void uctest_stack_pushPtr( void *ptr );

/**
 * @brief This returns the current stack pointer
 *
 * Returns the current stack pointer
 *
 * @return stack_t*
 */
stack_t *uctest_getStackPtr( void );

/**
 * @brief Set the current stack pointer
 *
 * Set the current stack pointer
 *
 * @param ptr The saved stack pointer to set
 */
void uctest_setStackPtr( stack_t *ptr );
#endif
#endif /** _UCTEST_UCTEST_DEVICE_H_ */

#ifdef __cplusplus
}
#endif
/** @} */
