#include <argp.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <simavr/sim_avr.h>
#include <simavr/sim_core.h>
#include <simavr/sim_core_decl.h>
#include <simavr/sim_elf.h>
#include <simavr/sim_gdb.h>

#include "test.h"

#include <libgen.h>

const char *argp_program_version = "1.0";
const char *argp_program_bug_address = "<sep@seplog.org>";
static char doc[] = "This program contains tests written using simavr to run them";
static char args_doc[] = "--mcu <device> <firmware>";

/**
 * We have to respect this regular expression.
 * /This program contains tests written using .*--(\w+)list_tests.*List the names of all tests
 * instead of running them/s
 */
static struct argp_option options[] = {
    { "freq", 'f', "Hz", 0, "Set the CPU frequency to <Hz>. Default: 8000000Hz" },
    { "mcu", 'm', "device", 0, "Simulate device <device>. See --list-cores for a complete list" },
    { "gdb", 'g', "port", OPTION_ARG_OPTIONAL, "Start a gdb server at <port>. Default: 1212" },
    { "list-cores", 0x80, 0, 0, "List all supported AVR cores and exit" },
    { "filter", 0x81, "PATTERNS", 0, "Filter for the test names" },
    { "avr_filter", 0x81, "PATTERNS", OPTION_ALIAS, NULL },
    { "list-tests", 0x82, NULL, 0, "List the names of all tests instead of running them" },
    { "avr_list_tests", 0x82, NULL, OPTION_ALIAS,
      "List the names of all tests instead of running them" },
    { "output", 0x83, "format:file", 0, "The output format and file" },
    { "avr_output", 0x83, NULL, OPTION_ALIAS, NULL },
    { "color", 0x84, "OPT", 0, NULL },
    { "avr_color", 0x84, "OPT", OPTION_HIDDEN, NULL },
    { "avr_also_run_disabled_tests", 0x90, NULL, OPTION_HIDDEN, NULL },
    { 0 } };

enum OUTPUT_FORMAT { OUTPUT_NONE, OUTPUT_XML };

struct arguments {
    unsigned int freq;
    char *mcu;
    char *firmware;
    bool gdb;
    unsigned short port;
    char *test_filter;
    bool test_list;
    enum test_output_format format;
    char *output;
    bool color;
};

static void list_cores() {
    printf( "Supported AVR cores:\n" );
    for( int i = 0; avr_kind[ i ]; i++ ) {
        printf( "       " );
        for( int ti = 0; ti < 4 && avr_kind[ i ]->names[ ti ]; ti++ )
            printf( "%s ", avr_kind[ i ]->names[ ti ] );
        printf( "\n" );
    }
}

static bool core_supported( const char *name ) {
    for( int i = 0; avr_kind[ i ]; i++ )
        for( int ti = 0; ti < 4 && avr_kind[ i ]->names[ ti ]; ti++ )
            if( strncmp( avr_kind[ i ]->names[ ti ], name, 48 ) == 0 )
                return true;
    return false;
}

static error_t parse_opt( int key, char *arg, struct argp_state *state ) {
    struct arguments *arguments = state->input;
    switch( key ) {
        case 'f': arguments->freq = atoi( arg ); break;
        case 'm':
            if( core_supported( arg ) )
                arguments->mcu = arg;
            else {
                argp_error( state, "Device not supported! Use --list-cores for supported devices" );
            }
            break;
        case 'g':
            arguments->gdb = true;
            if( arg != NULL )
                arguments->port = atoi( arg );
            break;
        case 0x81: arguments->test_filter = arg; break;
        case 0x82: arguments->test_list = true; break;
        case 0x83: {
            int len = strnlen( arg, BUFSIZ );
            char *colon = memchr( arg, ':', BUFSIZ );
            if( colon == NULL ) {
                arguments->output = arg;
            }
            else {
                if( strncmp( arg, "xml", ( colon - arg ) ) == 0 ) {
                    arguments->format = TEST_OUTPUT_XML;
                }
                arguments->output = ( colon + 1 );
            }
        } break;
        case 0x84:
            if( strncmp( arg, "no", 3 ) == 0 ) {
                arguments->color = false;
            }
            else if( strncmp( arg, "false", 6 ) == 0 ) {
                arguments->color = false;
            }
            break;
        case ARGP_KEY_ARG:
            switch( state->arg_num ) {
                case 0: arguments->firmware = arg; break;
                default: argp_usage( state ); break;
            }
            break;
        case 0x90: break;
        case ARGP_KEY_END:
            if( access( arguments->firmware, F_OK ) != 0 )
                argp_error( state, "File %s does not exist.", arguments->firmware );
            if( access( arguments->firmware, R_OK ) != 0 )
                argp_error( state, "File %s is not readable.", arguments->firmware );
            break;
        default: return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static struct argp argp = { options, parse_opt, args_doc, doc, 0, 0, 0 };

int main( int argc, char *argv[], char **envp ) {
    avr_t *avr;

    // char buffer[ BUFSIZ ];
    // snprintf( &buffer[ 0 ], BUFSIZ, "/tmp/%s.args", basename( argv[ 0 ] ) );
    // FILE *tmp_f = fopen( &buffer[ 0 ], "w" );
    // if( NULL == tmp_f ) {
    //     exit( EXIT_FAILURE );
    // }
    // for( int i = 0; i < argc; i++ ) {
    //     fprintf( tmp_f, "%s\n", argv[ i ] );
    // }
    // fclose( tmp_f );
    // tmp_f = NULL;

    char filenameBuffer[ 256 ];
    char *filename = NULL;
    snprintf( &filenameBuffer[ 0 ], 256, "%s.%s", argv[ 0 ], "elf" );
    if( access( &filenameBuffer[ 0 ], F_OK ) == 0 && access( &filenameBuffer[ 0 ], R_OK ) == 0 ) {
        filename = &filenameBuffer[ 0 ];
    }

    struct arguments arguments = { .freq = 0U, .port = 1212, .firmware = filename, .color = true };
    elf_firmware_t f = { { 0 } };

    arguments.mcu = getenv( "TEST_MCU" );
    char *freq = getenv( "TEST_FREQ" );
    if( freq != NULL ) {
        arguments.freq = atoi( freq );
    }

    if( !getenv( "ARGP_HELP_FMT" ) ) {
        putenv( "ARGP_HELP_FMT=rmargin=100" );
    }

    argp_parse( &argp, argc, argv, 0, 0, &arguments );

    int readelf_result = elf_read_firmware( arguments.firmware, &f );
    if( readelf_result == -1 ) {
        fprintf( stderr, "Could not read firmware file %s.\n", arguments.firmware );
        exit( EXIT_FAILURE );
    }

    if( arguments.mcu != NULL && strlen( arguments.mcu ) ) {
        strncpy( f.mmcu, arguments.mcu, 64 );
    }
    if( arguments.freq ) {
        f.frequency = arguments.freq;
    }

    avr = avr_make_mcu_by_name( f.mmcu );
    if( !avr ) {
        fprintf( stderr, "Could not create AVR device.\n" );
        exit( EXIT_FAILURE );
    }

    avr_init( avr );
    avr_load_firmware( avr, &f );

    test_t test = { .opt = { .filter = arguments.test_filter,
                             .firmware = arguments.firmware,
                             .output = { .format = arguments.format,
                                         .filename = arguments.output,
                                         .listTests = arguments.test_list,
                                         .color = arguments.color } },
                    .avr = avr,
                    .fw = &f };

    test_init( &test );

    avr->gdb_port = arguments.port;
    if( arguments.gdb ) {
        avr->state = cpu_Stopped;
        avr_gdb_init( avr );
    }

    for( ;; ) {
        int state = avr_run( avr );
        if( state == cpu_Done || state == cpu_Crashed )
            break;
    }

    avr_terminate( avr );

    return test.exitCode;
}
