#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <simavr/sim_avr.h>

#include "../src/commands.h"
#include "test.h"

#define TESTS_MAX ( 128 )

enum uctest_tst_flags {
    UCTEST_TEST_NONE,
    UCTEST_TEST_SIMPLE,
    UCTEST_TEST_FIXTURE,
    UCTEST_TEST_SETUP,
    UCTEST_TEST_TEARDOWN,
    UCTEST_TEST_SUCCESS,
    UCTEST_TEST_FAILED
};

struct uctest {
    avr_flashaddr_t addr;
    char *suite;
    char *name;
    enum uctest_tst_flags flags;
    char *file;
    uint16_t line;
    int messages_count;
    char **messages;
    avr_cycle_count_t cycles;
    time_t time;
    void *args;
};

struct uctest_suite {
    char *suite;
    int numTests;
    struct uctest *tests;
    int testsActive;
    int testsFailed;
    avr_cycle_count_t cycles;
    time_t time;
};

static inline uint16_t avr_read_word( uint8_t *buffer, uint16_t i ) {
    return ( uint16_t ) ( ( buffer[ i + 1 ] << 8 ) | buffer[ i ] );
}

static inline void avr_write_word( uint8_t *buffer, uint16_t i, uint16_t value ) {
    buffer[ i + 1 ] = value >> 8;
    buffer[ i ] = value & 0xFF;
}

static bool test_strnfilter( const char *str, const char *filter, size_t n ) {
    char *tst_str = ( char * ) str;
    char *tst_filter = ( char * ) filter;
    for( int i = 0; *tst_str && *tst_filter && ( i < n ); tst_str++, i++ ) {
        switch( *tst_filter ) {
            case '*':
                if( *tst_str == *( tst_filter + 1 ) ) {
                    if( test_strnfilter( ( const char * ) tst_str,
                                         ( const char * ) ( tst_filter + 1 ), n ) ) {
                        return true;
                    }
                }
                break;
            case '?': tst_filter++; break;
            default:
                if( *tst_filter == *tst_str )
                    tst_filter++;
                else
                    return false;
                break;
        }
    }
    return *tst_str == 0 &&
           ( *tst_filter == 0 || ( *tst_filter == '*' && *( tst_filter + 1 ) == 0 ) );
}

static bool test_strcmp( const char *name, const char *cmp ) {
    if( cmp == NULL )
        return true;
    char cmp_buffer[ BUFSIZ ];
    do {
        int len = strnlen( cmp, BUFSIZ );
        char *colon = memchr( cmp, ':', len );
        char *tmp_cmp = &cmp_buffer[ 0 ];
        if( NULL == colon ) {
            strncpy( tmp_cmp, cmp, BUFSIZ );
            cmp = ( cmp + len );
        }
        else {
            strncpy( tmp_cmp, cmp, ( colon - cmp ) );
            tmp_cmp[ ( colon - cmp ) ] = 0;
            cmp = colon + 1;
        }

        if( test_strnfilter( name, tmp_cmp, 48 ) ) {
            return true;
        }

    } while( *cmp );
    return false;
}

static void test_readFromFlash( avr_t *avr, avr_flashaddr_t addr, struct uctest *tst ) {
    tst->addr = addr;
    tst->suite = ( char * ) &avr->flash[ avr_read_word( avr->flash, addr + 0 ) ];
    tst->name = ( char * ) &avr->flash[ avr_read_word( avr->flash, addr + 2 ) ];
    tst->flags = avr_read_word( avr->flash, addr + 4 );
    tst->file = ( char * ) &avr->flash[ avr_read_word( avr->flash, addr + 6 ) ];
    tst->line = avr_read_word( avr->flash, addr + 8 );
    tst->cycles = 0;
    tst->messages_count = 0;
    tst->messages = NULL;
}

static void test_iterate( avr_t *avr, void ( *what )( avr_t *, struct uctest *, void * ),
                          void *args, avr_flashaddr_t *tests, int tests_count ) {
    for( int i = 0; i < tests_count; i++ ) {
        avr_flashaddr_t addr = tests[ i ];
        struct uctest test;
        test_readFromFlash( avr, addr, &test );
        // search for all tests also the disabled!
        if( test.flags == UCTEST_TEST_SIMPLE || test.flags == UCTEST_TEST_FIXTURE ||
            test.flags == UCTEST_TEST_NONE ) {
            what( avr, &test, args );
        }
    }
}

static void addSuite( avr_t *avr, struct uctest *tst, void *args ) {
    struct uctest_suite *suites = args;
    for( int i = 0; i < 128; i++ ) {
        if( NULL == suites[ i ].suite ) {
            suites[ i ].suite = tst->suite;
            break;
        }
        else if( strncmp( suites[ i ].suite, tst->suite, 128 ) == 0 ) {
            break;
        }
    }
}

struct getTest_struct {
    struct uctest_suite *suite;
    struct uctest *buffer;
};

static void getTestsForSuite( avr_t *t, struct uctest *tst, void *args ) {
    struct getTest_struct *arg = args;
    if( strncmp( arg->suite->suite, tst->suite, 128 ) == 0 ) {
        memcpy( &arg->buffer[ arg->suite->numTests ], tst, sizeof( struct uctest ) );
        arg->suite->numTests++;
    }
}

static int collectTests( avr_t *avr, struct uctest_suite *suites, int num,
                         avr_flashaddr_t *tests_ptr, int tests_count ) {
    struct uctest tests[ 128 ];
    struct getTest_struct args = { .buffer = &tests[ 0 ] };
    int allTests = 0;
    for( int i = 0; i < num; i++ ) {
        memset( &tests[ 0 ], 0, sizeof( tests ) );
        args.suite = ( suites + i );
        args.suite->numTests = 0;
        test_iterate( avr, getTestsForSuite, &args, tests_ptr, tests_count );
        if( args.suite->numTests > 0 ) {
            args.suite->tests = malloc( args.suite->numTests * sizeof( struct uctest ) );
            memcpy( args.suite->tests, &tests[ 0 ],
                    args.suite->numTests * sizeof( struct uctest ) );
        }
        allTests += args.suite->numTests;
    }
    return allTests;
}

static struct uctest_suite *collectSuites( test_t *tst, avr_flashaddr_t *tests, int tests_count ) {
    struct uctest_suite suites[ 128 ];
    memset( &suites[ 0 ], 0, sizeof( suites ) );
    test_iterate( tst->avr, addSuite, &suites[ 0 ], tests, tests_count );

    for( tst->suites_count = 0; suites[ tst->suites_count ].suite && tst->suites_count < 128;
         ( tst->suites_count )++ )
        ;

    struct uctest_suite *res = malloc( sizeof( struct uctest_suite ) * tst->suites_count );
    memcpy( res, &suites[ 0 ], sizeof( struct uctest_suite ) * tst->suites_count );

    tst->tests_count = collectTests( tst->avr, res, tst->suites_count, tests, tests_count );

    return res;
}

static void freeSuites( struct uctest_suite **suites_ptr, int count ) {
    struct uctest_suite *suites = *suites_ptr;
    for( int i = 0; i < count; i++ ) {
        if( suites[ i ].tests != NULL ) {
            for( int k = 0; k < suites[ i ].numTests; k++ ) {
                if( suites[ i ].tests[ k ].flags == UCTEST_TEST_SUCCESS ||
                    suites[ i ].tests[ k ].flags == UCTEST_TEST_FAILED ) {
                    for( int m = 0; m < suites[ i ].tests[ k ].messages_count; m++ ) {
                        free( suites[ i ].tests[ k ].messages[ m ] );
                        suites[ i ].tests[ k ].messages[ m ] = NULL;
                    }
                    free( suites[ i ].tests[ k ].messages );
                    suites[ i ].tests[ k ].messages = NULL;
                }
            }
            free( suites[ i ].tests );
            suites[ i ].tests = NULL;
        }
    }
    free( *suites_ptr );
    *suites_ptr = NULL;
}

struct list_args {
    char **current_suite;
    struct arguments *args;
};

static void test_print( avr_t *avr, struct uctest *test, void *args ) {
    char **name_ptr = ( char ** ) args;
    if( *name_ptr == NULL || strncmp( *name_ptr, test->suite, 48 ) != 0 ) {
        *name_ptr = test->suite;
        printf( "%s.\n", test->suite );
    }
    printf( "  %s\n", test->name );
}

static void test_list( test_t *tst ) {
    FILE *f = NULL;
    if( tst->opt.output.format != TEST_OUTPUT_NONE ) {
        f = fopen( tst->opt.output.filename, "w" );
        if( f == NULL ) {
            exit( EXIT_FAILURE );
        }
    }

    if( tst->opt.output.format == TEST_OUTPUT_XML ) {
        fprintf( f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" );
        fprintf( f, "<testsuites tests=\"%d\" name=\"AllTests\">\n", tst->tests_active );
    }

    struct uctest_suite *suite = tst->suites;
    for( int i = 0; i < tst->suites_count; i++ ) {
        if( suite[ i ].testsActive == 0 ) {
            continue;
        }
        printf( "%s.\n", suite[ i ].suite );
        if( tst->opt.output.format == TEST_OUTPUT_XML ) {
            fprintf( f, "  <testsuite name=\"%s\" tests=\"%d\">\n", suite[ i ].suite,
                     suite[ i ].numTests );
        }

        for( int k = 0; k < suite[ i ].numTests; k++ ) {
            printf( "  %s\n", suite[ i ].tests[ k ].name );
            if( tst->opt.output.format == TEST_OUTPUT_XML ) {
                fprintf( f, "    <testcase name=\"%s\" file=\"%s\" line=\"%d\" />\n",
                         suite[ i ].tests[ k ].name, suite[ i ].tests[ k ].file,
                         suite[ i ].tests[ k ].line );
            }
        }

        if( tst->opt.output.format == TEST_OUTPUT_XML ) {
            fprintf( f, "  </testsuite>\n" );
        }
    }

    if( tst->opt.output.format == TEST_OUTPUT_XML ) {
        fprintf( f, "</testsuites>\n" );
    }
    if( NULL != f ) {
        fclose( f );
    }
}

static void test_disable( test_t *tst ) {
    struct uctest_suite *suites = tst->suites;
    char name[ BUFSIZ ];
    tst->tests_active = 0;
    tst->suites_active = 0;
    for( int i = 0; i < tst->suites_count; i++ ) {
        suites[ i ].testsActive = 0;
        suites[ i ].testsFailed = 0;
        for( int k = 0; k < suites[ i ].numTests; k++ ) {
            snprintf( &name[ 0 ], BUFSIZ, "%s.%s", suites[ i ].suite, suites[ i ].tests[ k ].name );
            if( !test_strcmp( &name[ 0 ], tst->opt.filter ) ) {
                avr_write_word( tst->avr->flash, suites[ i ].tests[ k ].addr + 4, 0x00 );
            }
            else {
                tst->tests_active++;
                suites[ i ].testsActive++;
            }
        }
        if( suites[ i ].testsActive > 0 ) {
            tst->suites_active++;
        }
    }
}

static void callback_write_sexit( struct avr_t *avr, avr_io_addr_t addr, uint8_t v, void *param ) {
    int *exit_code = param;
    avr->state = ( !v ) ? cpu_Done : cpu_Crashed;
    *exit_code = v;
}

static inline char *getGreenColor( test_t *tst ) {
    return ( tst->opt.output.color ) ? "\e[32m" : "";
}

static inline char *getRedColor( test_t *tst ) {
    return ( tst->opt.output.color ) ? "\e[31m" : "";
}

static inline char *getColorEnd( test_t *tst ) {
    return ( tst->opt.output.color ) ? "\e[0m" : "";
}

static void printStartup( test_t *tst ) {
    printf( "Running main\n" );
    printf( "%s[==========]%s Running %d tests from %d test suites.\n", getGreenColor( tst ),
            getColorEnd( tst ), tst->tests_active, tst->suites_active );
}

static void printEnd( test_t *tst ) {
    struct uctest_suite *suites = tst->suites;

    printf( "%s[==========]%s %d tests from %d test suites ran. (%.0f ms total)\n",
            getGreenColor( tst ), getColorEnd( tst ), tst->tests_active, tst->suites_active,
            ( tst->cycles * 1000 ) / ( double ) tst->avr->frequency );

    printf( "%s[  PASSED  ]%s %d tests.\n", getGreenColor( tst ), getColorEnd( tst ),
            tst->tests_active - tst->tests_failed );

    if( tst->tests_failed > 0 ) {
        printf( "%s[  FAILED  ]%s %d test, listed below:\n", getRedColor( tst ), getColorEnd( tst ),
                tst->tests_failed );
        for( int i = 0; i < tst->suites_count; i++ ) {
            for( int k = 0; k < suites[ i ].numTests; k++ ) {
                if( suites[ i ].tests[ k ].flags == UCTEST_TEST_FAILED ) {
                    printf( "%s[  FAILED  ]%s %s.%s\n", getRedColor( tst ), getColorEnd( tst ),
                            suites[ i ].suite, suites[ i ].tests[ k ].name );
                }
            }
        }

        printf( "\n %d FAILED TEST\n", tst->tests_failed );
    }
}

static void printGlobalSetup( test_t *tst ) {
    printf( "%s[----------]%s Global test environment set-up.\n", getGreenColor( tst ),
            getColorEnd( tst ) );
}

static void printGlobalTeardown( test_t *tst ) {
    printf( "%s[----------]%s Global test environment tear-down.\n", getGreenColor( tst ),
            getColorEnd( tst ) );
}

static void printSuiteStatsStart( test_t *tst, struct uctest_suite *suite ) {
    printf( "%s[----------]%s %d tests from %s\n", getGreenColor( tst ), getColorEnd( tst ),
            suite->testsActive, suite->suite );
}

static void printSuiteStatsEnd( test_t *tst, struct uctest_suite *suite ) {
    printf( "%s[----------]%s %d tests from %s (%.0f ms total)\n\n", getGreenColor( tst ),
            getColorEnd( tst ), suite->testsActive, suite->suite,
            ( suite->cycles * 1000 ) / ( double ) tst->avr->frequency );
}

static void printTestBegin( test_t *tst, struct uctest *test ) {
    printf( "%s[ RUN      ]%s %s.%s\n", getGreenColor( tst ), getColorEnd( tst ), test->suite,
            test->name );
}

static void printTestSuccess( test_t *tst, struct uctest *test ) {
    printf( "%s[       OK ]%s %s.%s (%.0f ms)\n", getGreenColor( tst ), getColorEnd( tst ),
            test->suite, test->name,
            round( ( test->cycles * 1000 ) / ( double ) tst->avr->frequency ) );
}

static void printTestFailure( test_t *tst, struct uctest *test ) {
    if( test != NULL ) {
        printf( "%s[  FAILED  ]%s %s.%s (%.0f ms)\n", getRedColor( tst ), getColorEnd( tst ),
                test->suite, test->name,
                round( ( test->cycles * 1000 ) / ( double ) tst->avr->frequency ) );
    }
}

static void test_writeExcapeXml( FILE *f, char *str ) {
    while( *str ) {
        switch( *str ) {
            case '\n': fprintf( f, "%s", "&#x0A;" ); break;
            case '\"': fprintf( f, "%s", "&quot;" ); break;
            case '\'': fprintf( f, "%s", "&apos;" ); break;
            case '<': fprintf( f, "%s", "&lt;" ); break;
            case '>': fprintf( f, "%s", "&gt;" ); break;
            case '&': fprintf( f, "%s", "&amp;" ); break;
            default: fputc( *str, f ); break;
        }
        str++;
    }
}

static void test_makeOutput( test_t *tst ) {
    if( tst->opt.output.format == TEST_OUTPUT_NONE ) {
        return;
    }

    FILE *f = fopen( tst->opt.output.filename, "w" );
    if( NULL == f ) {
        exit( EXIT_FAILURE );
    }

    char all_buffer[ 24 ];
    struct tm *all_tm = localtime( &tst->time );
    strftime( &all_buffer[ 0 ], sizeof( all_buffer ), "%Y-%m-%dT%H:%M:%S", all_tm );

    fprintf( f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" );
    fprintf( f,
             "<testsuites tests=\"%d\" failures=\"%d\" disabled=\"0\" errors=\"0\" time=\"%.3f\" "
             "timestamp=\"%s\" name=\"AllTests\">\n",
             tst->tests_active, tst->tests_failed, tst->cycles / ( double ) tst->avr->frequency,
             &all_buffer[ 0 ] );

    struct uctest_suite *suites = tst->suites;
    for( int i = 0; i < tst->suites_count; i++ ) {
        struct uctest_suite *suite = &suites[ i ];
        if( !suites->testsActive ) {
            continue;
        }

        char suite_buffer[ 24 ];
        struct tm *suite_tm = localtime( &suite->time );
        strftime( &suite_buffer[ 0 ], sizeof( suite_buffer ), "%Y-%m-%dT%H:%M:%S", suite_tm );

        fprintf( f,
                 "  <testsuite name=\"%s\" tests=\"%d\" failures=\"%d\" disabled=\"0\" "
                 "skipped=\"0\" errors=\"0\" time=\"%.3f\" timestamp=\"%s\">\n",
                 suite->suite, suite->testsActive, suite->testsFailed,
                 suite->cycles / ( double ) tst->avr->frequency, &suite_buffer[ 0 ] );

        for( int k = 0; k < suite->numTests; k++ ) {
            struct uctest *test = &suite->tests[ k ];
            if( !( test->flags == UCTEST_TEST_FAILED || test->flags == UCTEST_TEST_SUCCESS ) ) {
                continue;
            }

            char test_buffer[ 24 ];
            struct tm *test_tm = localtime( &test->time );
            strftime( &test_buffer[ 0 ], sizeof( test_buffer ), "%Y-%m-%dT%H:%M:%S", test_tm );

            fprintf( f,
                     "    <testcase name=\"%s\" status=\"run\" result=\"completed\" "
                     "time=\"%.3f\" timestamp=\"%s\" classname=\"%s\"",
                     test->name, test->cycles / ( double ) tst->avr->frequency, &test_buffer[ 0 ],
                     test->suite );
            if( !test->messages_count ) {
                fprintf( f, " />\n" );
            }
            else {
                fprintf( f, ">\n" );

                for( int m = 0; m < test->messages_count; m++ ) {
                    fprintf( f, "      <failure message=\"" );
                    test_writeExcapeXml( f, test->messages[ m ] );
                    fprintf( f, "\" type=\"\"><![CDATA[%s]]></failure>\n", test->messages[ m ] );
                }

                fprintf( f, "    </testcase>\n" );
            }
        }

        fprintf( f, "  </testsuite>\n" );
    }

    fprintf( f, "</testsuites>\n" );
    fclose( f );
}

void *test_start_hook( avr_t *avr, char * ) __attribute__( ( weak ) );

void test_end_hook( avr_t *, char *, uint8_t, void * ) __attribute__( ( weak ) );

static int test_prepare( test_t *tst, avr_flashaddr_t *tests, int tests_count ) {
    tst->suites = collectSuites( tst, tests, tests_count );
    test_disable( tst );
    if( tst->opt.output.listTests ) {
        test_list( tst );
        tst->avr->state = cpu_Done;
        freeSuites( ( struct uctest_suite ** ) &tst->suites, tst->suites_count );
        return 1;
    }
    return 0;
}

static void callback_write( struct avr_t *avr, avr_io_addr_t addr, uint8_t v, void *param ) {
    static int i = 0;

    static struct uctest_suite *current_suite;
    static struct uctest *current_test;

    static uint8_t buffer[ BUFSIZ ];
    static int buffer_index = 0;
    static int buffer_length = 0;
    static int buffer_next = 0;
    static bool readString = false;

    static avr_flashaddr_t test_buffer[ TESTS_MAX ];
    static int test_buffer_count = 0;

    static avr_cycle_count_t cycles_start;

    if( buffer_length > 0 && buffer_index < buffer_length ) {
        buffer[ buffer_index++ ] = v;
        if( buffer_index == buffer_length || ( readString == true && v == 0 ) ) {
            v = buffer_next;
        }
        else {
            return;
        }
    }

    test_t *tst = ( test_t * ) param;
    switch( v ) {
        case SIM_CMD_START:
            if( test_prepare( tst, &test_buffer[ 0 ], test_buffer_count ) == 1 ) {
                avr->state = cpu_Done;
                return;
            }
            else {
                printStartup( tst );
                buffer_length = 0;
                tst->tests_failed = 0;
                tst->cycles = 0;
            }
            break;
        case SIM_CMD_END:
            printEnd( tst );
            test_makeOutput( tst );
            freeSuites( ( struct uctest_suite ** ) &tst->suites, tst->suites_count );
            break;
        case SIM_CMD_GLOBAL_SETUP: printGlobalSetup( tst ); break;
        case SIM_CMD_GLOBAL_TEARDOWN: printGlobalTeardown( tst ); break;
        case SIM_CMD_TEST_REGISTER:
            if( buffer_length == 0 ) {
                buffer_length = 2;
                buffer_index = 0;
                buffer_next = SIM_CMD_TEST_REGISTER;
            }
            else {
                buffer_length = 0;
                buffer_index = 0;
                test_buffer[ test_buffer_count ] = ( ( buffer[ 0 ] << 8 ) | buffer[ 1 ] );
                struct uctest current;
                test_readFromFlash( avr, test_buffer[ test_buffer_count ], &current );
                // We have to disable this now otherwise the controller runs the test!
                char name[ BUFSIZ ];
                snprintf( &name[ 0 ], BUFSIZ, "%s.%s", current.suite, current.name );
                if( current.flags == UCTEST_TEST_SIMPLE || current.flags == UCTEST_TEST_FIXTURE ) {
                    if( !test_strcmp( &name[ 0 ], tst->opt.filter ) ) {
                        avr_write_word( tst->avr->flash, current.addr + 4, 0x00 );
                    }
                }
                test_buffer_count++;
            }
            break;
        case SIM_CMD_SUITE_START:
            if( buffer_length == 0 ) {
                buffer_length = 4;
                buffer_index = 0;
                buffer_next = SIM_CMD_SUITE_START;
            }
            else {
                buffer_length = 0;
                buffer_index = 0;
                char *suitename = ( char * ) &avr->flash[ ( buffer[ 2 ] << 8 ) | buffer[ 3 ] ];
                int number = ( buffer[ 0 ] << 8 ) | buffer[ 1 ];
                current_suite = NULL;
                for( int i = 0; i < tst->suites_count; i++ ) {
                    struct uctest_suite *suites = tst->suites;
                    if( strncmp( suitename, suites[ i ].suite, 48 ) == 0 ) {
                        current_suite = &suites[ i ];
                        break;
                    }
                }
                if( current_suite == NULL ) {
                    exit( 1 );
                }
                tst->time = time( NULL );
                current_suite->time = time( NULL );
                current_suite->cycles = 0;
                printSuiteStatsStart( tst, current_suite );
            }
            break;
        case SIM_CMD_SUITE_END:
            printSuiteStatsEnd( tst, current_suite );
            current_suite = NULL;
            current_test = NULL;
            break;
        case SIM_CMD_TEST_START:
            if( buffer_length == 0 ) {
                buffer_index = 0;
                buffer_length = 2;
                buffer_next = SIM_CMD_TEST_START;
            }
            else {
                buffer_index = 0;
                buffer_length = 0;
                char *tstname = ( char * ) &avr->flash[ ( buffer[ 0 ] << 8 ) | buffer[ 1 ] ];
                current_test = NULL;
                for( int i = 0; i < current_suite->numTests; i++ ) {
                    if( strncmp( tstname, current_suite->tests[ i ].name, 48 ) == 0 ) {
                        current_test = &current_suite->tests[ i ];
                    }
                }
                if( current_test == NULL ) {
                    exit( 1 );
                }
                current_test->time = time( NULL );
                current_test->cycles = 0;
                current_test->flags = UCTEST_TEST_SUCCESS;
                current_test->messages_count = 0;
                current_test->messages = malloc( sizeof( char * ) * 128 );
                printTestBegin( tst, current_test );
                if( test_start_hook ) {
                    char name[ 128 ];
                    snprintf( &name[ 0 ], 128, "%s.%s", current_suite->suite, current_test->name );
                    current_test->args = test_start_hook( avr, &name[ 0 ] );
                }
            }
            break;
        case SIM_CMD_CLK_START: cycles_start = tst->avr->cycle; break;
        case SIM_CMD_CLK_END: current_test->cycles += tst->avr->cycle - cycles_start; break;
        case SIM_CMD_TEST_SUCCESS:
            printTestSuccess( tst, current_test );
            tst->cycles += current_test->cycles;
            current_suite->cycles += current_test->cycles;
            if( test_end_hook ) {
                char name[ 128 ];
                snprintf( &name[ 0 ], 128, "%s.%s", current_suite->suite, current_test->name );
                test_end_hook( avr, &name[ 0 ], 0, current_test->args );
            }
            current_test = NULL;
            break;
        case SIM_CMD_TEST_FAILURE:
            if( current_test != NULL ) {
                printTestFailure( tst, current_test );
                tst->tests_failed++;
                tst->cycles += current_test->cycles;
                current_suite->cycles += current_test->cycles;
                if( test_end_hook ) {
                    char name[ 128 ];
                    snprintf( &name[ 0 ], 128, "%s.%s", current_suite->suite, current_test->name );
                    test_end_hook( avr, &name[ 0 ], 1, current_test->args );
                }
                current_test = NULL;
                if( current_suite != NULL ) {
                    current_suite->testsFailed++;
                }
            }
            break;
        case SIM_CMD_TEST_MSG_FAILURE:
            if( buffer_index == 0 ) {
                buffer_length = BUFSIZ;
                readString = true;
                buffer_next = SIM_CMD_TEST_MSG_FAILURE;
                current_test->cycles += tst->avr->cycle - cycles_start;
            }
            else {
                // Make shure that we do not read the next string
                buffer_length = buffer_index;
                buffer[ buffer_index - 1 ] = '#';
                readString = false;
                current_test->flags = UCTEST_TEST_FAILED;
            }
            break;
        case SIM_CMD_TEST_MSG:
            if( buffer_length == buffer_index ) {
                buffer_length = BUFSIZ;
                readString = true;
                buffer_next = SIM_CMD_TEST_MSG;
            }
            else {
                char *str_failure = malloc( buffer_index );
                memset( str_failure, 0, buffer_index );
                char *tmp_splitter = memchr( &buffer[ 0 ], '#', buffer_index );

                memcpy( str_failure, &buffer[ 0 ], ( tmp_splitter - ( char * ) &buffer[ 0 ] ) );
                *( str_failure + ( tmp_splitter - ( char * ) &buffer[ 0 ] ) + 1 ) = 0;
                printf( "%s: Failure\n", str_failure );
                printf( "%s", tmp_splitter + 1 );

                *tmp_splitter = '\n';
                memcpy( str_failure, &buffer[ 0 ], buffer_index );

                current_test->messages[ current_test->messages_count++ ] = str_failure;

                buffer_index = 0;
                buffer_length = 0;
                readString = false;
                cycles_start = tst->avr->cycle;
            }
            break;
        default: break;
    }
}

void test_init_hook( avr_t *avr ) __attribute__( ( weak ) );

void test_init( test_t *tst ) {
    tst->exitCode = EXIT_SUCCESS;
    avr_register_io_write( tst->avr, 0x21, callback_write_sexit, &tst->exitCode );
    avr_register_io_write( tst->avr, 0x20, callback_write, tst );

    if( test_init_hook ) {
        test_init_hook( tst->avr );
    }
}
