/**
 * @defgroup   sim Simulation: A simulator for the uCTest Framework
 *
 * This is a simulator for the uCTestframework to make this GTest compatible.
 *
 * @author     Sebastian Koschmieder <sep@seplog.org>
 * @date       2021-05-25 12:35:29
 * @{
 */

#if !defined( _UCTEST_SIMULATOR_TEST_H_ )
#ifdef __cplusplus
extern "C" {
#endif

#ifndef _UCTEST_SIMULATOR_TEST_H_
#define _UCTEST_SIMULATOR_TEST_H_
#endif

#include <stdbool.h>
#include <time.h>

#include <simavr/sim_avr.h>
#include <simavr/sim_elf.h>

#if !defined( __DOXYGEN__ )

enum test_output_format { TEST_OUTPUT_NONE, TEST_OUTPUT_XML };

struct test_output {
    enum test_output_format format;
    const char *filename;
    bool listTests;
    bool color;
};

struct test_options {
    const char *firmware;
    const char *filter;
    struct test_output output;
};

typedef struct {
    avr_t *avr;
    elf_firmware_t *fw;
    struct test_options opt;

    int suites_count;
    int tests_count;
    int suites_active;
    int tests_active;
    int tests_failed;
    avr_cycle_count_t cycles;
    void *suites;
    int exitCode;
    time_t time;
} test_t;

extern void test_init( test_t * );
#endif

#if defined( __DOXYGEN__ )

/**
 * @brief      This function is executed when the simulator starts
 *
 * This function is executed when the simulator starts. Define the hook to initialize the hardware
 * of the simulation.
 *
 * @param[in]  avr An instance of the avr_t struct
 */
void test_init_hook( avr_t *avr );

/**
 * @brief      This hook is called when a test starts its execution
 *
 * This hook is called when a test starts its execution. It is possible to use this as a start
 * signal for an FSM for a special hardware simulation.
 *
 * @param[in]  avr An instance of the avr_t struct
 * @param[in]  name The name of the test. In format "suitename.testname"
 * @return some userdata or parameter
 */
void *test_start_hook( avr_t *avr, char *name );

/**
 * @brief      Called when the test is finished
 *
 * This hook is called when the test is finished. The params are a pointer from the return value of
 * the start hook.
 *
 * @param[in]  avr An instance of the avr_t struct
 * @param[in]  name The name of the test. In format "suitename.testname"
 * @param[in]  result The test result. 0 if the test was successful and 1 if the test faild
 * @param[in]  param The user parameter created in the start hook
 */
void test_end_hook( avr_t *avr, char *name, uint8_t result, void *param );

#endif
#endif /** _UCTEST_SIMULATOR_TEST_H_ */

#ifdef __cplusplus
}
#endif
/** @} */
